/* Copyright 2016 The TensorFlow Authors. All Rights Reserved.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
==============================================================================*/

#include <string.h>
#include <memory>

#include "tensorflow/c/c_api.h"
#include "tensorflow/c/c_api_internal.h"
#include "tensorflow/java/src/main/native/utils_jni.h"
#include "tensorflow/java/src/main/native/exception_jni.h"
#include "tensorflow/java/src/main/native/session_jni.h"

#include "tensorflow/core/common_runtime/direct_session.h"
//#include "tensorflow/cc/client/client_session.h"

#include "cuda_runtime_api.h"

namespace {
TF_Session* requireHandle(JNIEnv* env, jlong handle) {
  static_assert(sizeof(jlong) >= sizeof(TF_Session*),
                "Cannot package C object pointers as a Java long");
  if (handle == 0) {
    throwException(env, kNullPointerException,
                   "close() has been called on the Session");
    return nullptr;
  }
  return reinterpret_cast<TF_Session*>(handle);
}

template <class T>
void resolveHandles(JNIEnv* env, const char* type, jlongArray src_array,
                    T** dst, jint n) {
  if (env->ExceptionCheck()) return;
  jint len = env->GetArrayLength(src_array);
  if (len != n) {
    throwException(env, kIllegalArgumentException, "expected %d, got %d %s", n,
                   len, type);
    return;
  }
  jlong* src_start = env->GetLongArrayElements(src_array, nullptr);
  jlong* src = src_start;
  for (int i = 0; i < n; ++i, ++src, ++dst) {
    if (*src == 0) {
      throwException(env, kNullPointerException, "invalid %s (#%d of %d)", type,
                     i, n);
      break;
    }
    *dst = reinterpret_cast<T*>(*src);
  }
  env->ReleaseLongArrayElements(src_array, src_start, JNI_ABORT);
}

void TF_MaybeDeleteBuffer(TF_Buffer* buf) {
  if (buf == nullptr) return;
  TF_DeleteBuffer(buf);
}

typedef std::unique_ptr<TF_Buffer, decltype(&TF_MaybeDeleteBuffer)>
    unique_tf_buffer;

unique_tf_buffer MakeUniqueBuffer(TF_Buffer* buf) {
  return unique_tf_buffer(buf, TF_MaybeDeleteBuffer);
}

}  // namespace

JNIEXPORT jlong JNICALL Java_org_tensorflow_Session_allocate(
    JNIEnv* env, jclass clazz, jlong graph_handle) {
  return Java_org_tensorflow_Session_allocate2(env, clazz, graph_handle,
                                               nullptr, nullptr);
}

JNIEXPORT jlong JNICALL Java_org_tensorflow_Session_allocate2(
    JNIEnv* env, jclass clazz, jlong graph_handle, jstring target,
    jbyteArray config) {
  if (graph_handle == 0) {
    throwException(env, kNullPointerException, "Graph has been close()d");
    return 0;
  }
  TF_Graph* graph = reinterpret_cast<TF_Graph*>(graph_handle);
  TF_Status* status = TF_NewStatus();
  TF_SessionOptions* opts = TF_NewSessionOptions();
  jbyte* cconfig = nullptr;
  if (config != nullptr) {
    cconfig = env->GetByteArrayElements(config, nullptr);
    TF_SetConfig(opts, cconfig,
                 static_cast<size_t>(env->GetArrayLength(config)), status);
    if (!throwExceptionIfNotOK(env, status)) {
      env->ReleaseByteArrayElements(config, cconfig, JNI_ABORT);
      TF_DeleteSessionOptions(opts);
      TF_DeleteStatus(status);
      return 0;
    }
  }
  const char* ctarget = nullptr;
  if (target != nullptr) {
    ctarget = env->GetStringUTFChars(target, nullptr);
  }
  TF_Session* session = TF_NewSession(graph, opts, status);
  if (config != nullptr) {
    env->ReleaseByteArrayElements(config, cconfig, JNI_ABORT);
  }
  if (target != nullptr) {
    env->ReleaseStringUTFChars(target, ctarget);
  }
  TF_DeleteSessionOptions(opts);
  bool ok = throwExceptionIfNotOK(env, status);
  TF_DeleteStatus(status);

  return ok ? reinterpret_cast<jlong>(session) : 0;
}

JNIEXPORT void JNICALL Java_org_tensorflow_Session_delete(JNIEnv* env,
                                                          jclass clazz,
                                                          jlong handle) {
  TF_Session* session = requireHandle(env, handle);
  if (session == nullptr) return;
  TF_Status* status = TF_NewStatus();
  TF_CloseSession(session, status);
  // Result of close is ignored, delete anyway.
  TF_DeleteSession(session, status);
  throwExceptionIfNotOK(env, status);
  TF_DeleteStatus(status);
}

JNIEXPORT jbyteArray JNICALL Java_org_tensorflow_Session_run(
    JNIEnv* env, jclass clazz, jlong handle, jbyteArray jrun_options,
    jlongArray input_tensor_handles, jlongArray input_op_handles,
    jintArray input_op_indices, jlongArray output_op_handles,
    jintArray output_op_indices, jlongArray target_op_handles,
    jboolean want_run_metadata, jlongArray output_tensor_handles) {
  TF_Session* session = requireHandle(env, handle);
  if (session == nullptr) return nullptr;

  const jint ninputs = env->GetArrayLength(input_tensor_handles);
  const jint noutputs = env->GetArrayLength(output_tensor_handles);
  const jint ntargets = env->GetArrayLength(target_op_handles);

  std::unique_ptr<TF_Output[]> inputs(new TF_Output[ninputs]);
  std::unique_ptr<TF_Tensor* []> input_values(new TF_Tensor*[ninputs]);
  std::unique_ptr<TF_Output[]> outputs(new TF_Output[noutputs]);
  std::unique_ptr<TF_Tensor* []> output_values(new TF_Tensor*[noutputs]);
  std::unique_ptr<TF_Operation* []> targets(new TF_Operation*[ntargets]);
  unique_tf_buffer run_metadata(
      MakeUniqueBuffer(want_run_metadata ? TF_NewBuffer() : nullptr));

  resolveHandles(env, "input Tensors", input_tensor_handles, input_values.get(),
                 ninputs);
  resolveOutputs(env, "input", input_op_handles, input_op_indices, inputs.get(),
                 ninputs);
  resolveOutputs(env, "output", output_op_handles, output_op_indices,
                 outputs.get(), noutputs);
  resolveHandles(env, "target Operations", target_op_handles, targets.get(),
                 ntargets);
  if (env->ExceptionCheck()) return nullptr;

  TF_Status* status = TF_NewStatus();

  unique_tf_buffer run_options(MakeUniqueBuffer(nullptr));
  jbyte* jrun_options_data = nullptr;
  if (jrun_options != nullptr) {
    size_t sz = env->GetArrayLength(jrun_options);
    if (sz > 0) {
      jrun_options_data = env->GetByteArrayElements(jrun_options, nullptr);
      run_options.reset(
          TF_NewBufferFromString(static_cast<void*>(jrun_options_data), sz));
    }
  }

  TF_SessionRun(session, run_options.get(), inputs.get(), input_values.get(),
                static_cast<int>(ninputs), outputs.get(), output_values.get(),
                static_cast<int>(noutputs), targets.get(),
                static_cast<int>(ntargets), run_metadata.get(), status);

  if (jrun_options_data != nullptr) {
    env->ReleaseByteArrayElements(jrun_options, jrun_options_data, JNI_ABORT);
  }

  if (!throwExceptionIfNotOK(env, status)) {
    TF_DeleteStatus(status);
    return nullptr;
  }
  jlong* t = env->GetLongArrayElements(output_tensor_handles, nullptr);
  for (int i = 0; i < noutputs; ++i) {
    t[i] = reinterpret_cast<jlong>(output_values[i]);
  }
  env->ReleaseLongArrayElements(output_tensor_handles, t, 0);

  jbyteArray ret = nullptr;
  if (run_metadata != nullptr) {
    ret = env->NewByteArray(run_metadata->length);
    env->SetByteArrayRegion(ret, 0, run_metadata->length,
                            reinterpret_cast<const jbyte*>(run_metadata->data));
  }
  TF_DeleteStatus(status);
  return ret;
}

// Create an empty tensor of type 'dtype'. 'shape' can be arbitrary, but has to
// result in a zero-sized tensor.
static TF_Tensor* EmptyTensor(TF_DataType dtype,
                              const tensorflow::TensorShape& shape) {
  static char empty;
  tensorflow::int64 nelems = 1;
  std::vector<tensorflow::int64> dims;
  for (int i = 0; i < shape.dims(); ++i) {
    dims.push_back(shape.dim_size(i));
    nelems *= shape.dim_size(i);
  }
  CHECK_EQ(nelems, 0);
  static_assert(sizeof(int64_t) == sizeof(tensorflow::int64),
                "64-bit int types should match in size");
  return TF_NewTensor(
      dtype, reinterpret_cast<const int64_t*>(dims.data()), shape.dims(),
      reinterpret_cast<void*>(&empty), 0, [](void*, size_t, void*) {}, nullptr);
}

JNIEXPORT jbyteArray JNICALL Java_org_tensorflow_Session_elphelRunCallable(
    JNIEnv* env, jclass clazz, 
	jlong session_handle, jlong callable_handle,
    jlongArray input_tensor_handles,
	jlongArray output_tensor_handles) {
	
	//printf("Running Callable\n");
	
	TF_Session* session = requireHandle(env, session_handle);
	
	using namespace tensorflow;
	
	Session::CallableHandle feed_gpu_fetch_cpu = (Session::CallableHandle) reinterpret_cast<long>(callable_handle);
	
    const jint ninputs = env->GetArrayLength(input_tensor_handles);
    const jint noutputs = env->GetArrayLength(output_tensor_handles);
	
    //printf("ninputs: %d, noutputs: %d\n",ninputs, noutputs);
    
    std::unique_ptr<TF_Tensor* []> output_values(new TF_Tensor*[noutputs]);
    std::unique_ptr<TF_Tensor* []> input_values(new TF_Tensor*[ninputs]);
    
    // from input tensor handles to inputs?
    resolveHandles(env, "input Tensors", input_tensor_handles, input_values.get(), ninputs);    
	
	std::vector<Tensor> inputs(ninputs);
	for (int i=0; i<ninputs; ++i) {
		TF_TensorToTensor(input_values[i],&inputs[i]);
	}
	
	// figure out how to create stuff from handles
    std::vector<Tensor> outputs(noutputs);
    
	auto runStatus = session->session->RunCallable(feed_gpu_fetch_cpu, {inputs}, &outputs, nullptr);
    
	if (!runStatus.ok()){
		printf("It is with a heavy heart I inform you that RunCallable has failed. Here's the error message:\n");
		printf(runStatus.error_message().c_str());
		return nullptr;
	}
	
	// get the handles t
    jlong* t = env->GetLongArrayElements(output_tensor_handles, nullptr);
    TF_Status* status = TF_NewStatus();
    
    for (int i = 0; i < noutputs; ++i) {
      //outputs[i] = inputz[i];
      const Tensor& src = outputs[i];
      
      /*
      std::cout << src.DebugString() << std::endl;
      // print values:
      std::cout << "Output tensor (printing from session_jni.cc):";
      auto tmap = src.tensor<float, 1>();
	  for (int d = 0; d < 256; d++){
		  std::cout << (int) tmap(d);
		  if (d!=255) std::cout << ", ";
	  }
	  */
	  
      //output_values[i]->tensor = outputs[i];
	  if (!src.IsInitialized() || src.NumElements() == 0) {
		  output_values[i] = EmptyTensor(static_cast<TF_DataType>(src.dtype()), src.shape());
	      continue;
	  }
	  output_values[i] = TF_TensorFromTensor(src, status);
	  // for whatever reason status cannot be a nullptr here
      //output_values[i] = TF_TensorFromTensor(src, nullptr);
      t[i] = reinterpret_cast<jlong>(output_values[i]);
    }
    // this copies back the updated array andit can be accessed up there in Java
    env->ReleaseLongArrayElements(output_tensor_handles, t, 0);

	jbyteArray ret = nullptr;
	return ret;
}

JNIEXPORT jstring JNICALL Java_org_tensorflow_Session_elphelGetGPUDeviceName(JNIEnv* env,
                                                          jclass clazz,
                                                          jlong handle) {
  TF_Session* session = requireHandle(env, handle);
  if (session == nullptr) return env->NewStringUTF("");

  using namespace tensorflow;

  std::vector<DeviceAttributes> devices;
  TF_CHECK_OK(session->session->ListDevices(&devices));

  for (const DeviceAttributes& d : devices) {
    LOG(INFO) << "Device: " << d.name();
    if (d.device_type() == "GPU" || d.device_type() == "gpu") {
      return env->NewStringUTF(d.name().c_str());
    }
  }

  return env->NewStringUTF("");
}

JNIEXPORT jlong JNICALL Java_org_tensorflow_Session_elphelMakeCallable(
    JNIEnv* env, jclass clazz, jlong session_handle, jbyteArray config){
	
	TF_Session* session = requireHandle(env, session_handle);
	
	using namespace tensorflow;
	
	CallableOptions opts;
	
	jbyte* cconfig = nullptr;
	if (config != nullptr) {
	  cconfig = env->GetByteArrayElements(config, nullptr);
	  opts.ParseFromArray(cconfig, static_cast<size_t>(env->GetArrayLength(config)));
	}
	
	Session::CallableHandle feed_gpu_fetch_cpu;
	
	auto runStatus = session->session->MakeCallable(opts, &feed_gpu_fetch_cpu);
	
	if (!runStatus.ok()){
		printf("It is with a heavy heart I inform you that MakeCallable has failed. Here's the error message:\n");
		printf(runStatus.error_message().c_str());
		return -1;
	}else{
		/*
		jlong* t = env->GetLongArrayElements(callable_handle, nullptr);
		t[0] = reinterpret_cast<jlong>((long) feed_gpu_fetch_cpu);
		env->ReleaseLongArrayElements(callable_handle, t, 0);
		*/
		return reinterpret_cast<jlong>((long) feed_gpu_fetch_cpu);
	}
}







